from datetime import date
from unittest import TestCase, skip

import numpy as np

from axisutilities import Axis, AxisConverter, DailyTimeAxisBuilder, WeeklyTimeAxisBuilder, \
    RollingWindowTimeAxisBuilder, MonthlyTimeAxisBuilder


class TestTimeAxisConverter(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        import os
        os.environ['TZ'] = 'MST'

    def test_get_coverage_01(self):
        from_axis = Axis(
            lower_bound=np.arange(0, 7)*24,
            upper_bound=np.arange(1, 8)*24,
            fraction=0.5
        )

        to_axis = Axis(
            lower_bound=[0],
            upper_bound=[168],
            data_ticks=[84]
        )

        row_idx, col_idx, weights = AxisConverter._get_coverage(
            from_axis.lower_bound, from_axis.upper_bound,
            to_axis.lower_bound, to_axis.upper_bound
        )

        self.assertListEqual([0]*7, row_idx)
        self.assertListEqual(list(range(7)), col_idx)
        self.assertListEqual([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0], weights)

    def test_get_coverage_02(self):
        from_axis = Axis(
            lower_bound=[0],
            upper_bound=[168],
            data_ticks=[84]
        )

        to_axis = Axis(
            lower_bound=np.arange(0, 7) * 24,
            upper_bound=np.arange(1, 8) * 24,
            fraction=0.5
        )

        row_idx, col_idx, coverageWeight = AxisConverter._get_coverage(
            from_axis.lower_bound, from_axis.upper_bound,
            to_axis.lower_bound, to_axis.upper_bound
        )
        self.assertListEqual(list(range(7)), row_idx)
        self.assertListEqual([0]*7, col_idx)
        self.assertListEqual([1.0/7.0, 1.0/7.0, 1.0/7.0, 1.0/7.0, 1.0/7.0, 1.0/7.0, 1.0/7.0], coverageWeight)

    def test_get_coverage_03(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        row_idx, col_idx, weights = AxisConverter._get_coverage(
            from_axis.lower_bound, from_axis.upper_bound,
            to_axis.lower_bound, to_axis.upper_bound
        )

        self.assertListEqual([0] * 7, row_idx)
        self.assertListEqual(list(range(7)), col_idx)
        self.assertListEqual([1.0] * 7, weights)

    def test_creation_01(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

    def test_average_01(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = [1, 2, 3, 4, 5, 6, 7]

        to_data = tc.average(from_data)

        self.assertAlmostEqual(4.0, to_data[0, 0], 0)

    def test_average_02(self):
        from_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        to_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = [4.0]

        to_data = tc.average(from_data)

        self.assertEqual((tc.to_nelem, 1), to_data.shape)

        for e in to_data:
            self.assertAlmostEqual(4.0, e[0], 0)

    def test_average_03(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_end_date(date(2019, 1, 8)) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.asarray([1, 2, 3, 4, 5, 6, 7])

        to_data = tc.average(from_data)

        self.assertAlmostEqual(4.0, to_data[0, 0], 0)

    def test_average_04(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.average(from_data)

        self.assertAlmostEqual(4.0, to_data[0, 0], 0)
        self.assertAlmostEqual(11.0, to_data[1, 0], 0)

    def test_average_05(self):
        from_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        to_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = [4.0, 11]

        to_data = tc.average(from_data)

        expected = [4] * 7 + [11] * 7
        self.assertListEqual(expected, to_data.flatten().tolist())

    def test_average_06(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 0)

        to_data = tc.average(from_data)

        self.assertEqual((2, 3, 4), to_data.shape)

        self.assertListEqual(
            [4] * 12,
            np.round(to_data[0, :, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

        self.assertListEqual(
            [11] * 12,
            np.round(to_data[1, :, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

    def test_average_07(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14))

        to_data = tc.average(from_data, time_dimension=2)

        self.assertEqual((3, 4, 2), to_data.shape)

        self.assertListEqual(
            [4] * 12,
            np.round(to_data[:, :, 0].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

        self.assertListEqual(
            [11] * 12,
            np.round(to_data[:, :, 1].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

    def test_average_08(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1)

        to_data = tc.average(from_data, time_dimension=1)

        self.assertEqual((3, 2, 4), to_data.shape)

        self.assertListEqual(
            [4] * 12,
            np.round(to_data[:, 0, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

        self.assertListEqual(
            [11] * 12,
            np.round(to_data[:, 1, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

    def test_average_09(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1).tolist()

        to_data = tc.average(from_data, time_dimension=1)

        self.assertEqual((3, 2, 4), to_data.shape)

        self.assertListEqual(
            [4] * 12,
            np.round(to_data[:, 0, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

        self.assertListEqual(
            [11] * 12,
            np.round(to_data[:, 1, :].reshape((-1, 1))).astype(np.int).flatten().tolist()
        )

    def test_average_10(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=3
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.average(from_data)

        self.assertAlmostEqual(4.0, to_data[0, 0], 0)
        self.assertAlmostEqual(11.0, to_data[1, 0], 0)
        self.assertTrue(np.isnan(to_data[2, 0]))

    def test_average_11(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = RollingWindowTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            end_date=date(2019, 1, 15),
            window_size=7
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.average(from_data)
        self.assertEqual(8, to_data.size)

        for idx in range(8):
            self.assertAlmostEqual(idx+4, to_data[idx, 0], 0)

    def test_average_12(self):
        from_axis = DailyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(14) \
            .build()

        to_axis = WeeklyTimeAxisBuilder()\
            .set_start_date(date(2019, 1, 1)) \
            .set_n_interval(2) \
            .build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))
        from_data[2] = np.nan

        to_data = tc.average(from_data)

        self.assertAlmostEqual(4.167, to_data[0, 0], 3)
        self.assertAlmostEqual(11.0, to_data[1, 0], 3)

    def test_min_01(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.min(from_data)

        self.assertAlmostEqual(1.0, to_data[0, 0], 0)
        self.assertAlmostEqual(8.0, to_data[1, 0], 0)

    def test_min_02(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 0).tolist()

        to_data = tc.min(from_data)

        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") - to_data[0].round() == 0.0
            )
        )
        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 8 - to_data[1].round() == 0.0
            )
        )

    def test_min_03(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1).tolist()

        to_data = tc.min(from_data, time_dimension=1)

        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") - to_data[:, 0, :].round() == 0.0
            )
        )
        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 8 - to_data[:, 1, :].round() == 0.0
            )
        )

    def test_min_04(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = RollingWindowTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            end_date=date(2019, 1, 15),
            window_size=7
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1).tolist()

        to_data = tc.min(from_data, time_dimension=1)

        for i in range(tc.to_nelem):
            self.assertTrue(
                np.all(
                    np.ones((3, 4), dtype="int") * (1 + i) - to_data[:, i, :].round() == 0.0
                )
            )

    def test_min_05(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=3
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.min(from_data)

        self.assertAlmostEqual(1.0, to_data[0, 0], 0)
        self.assertAlmostEqual(8.0, to_data[1, 0], 0)
        self.assertTrue(np.isnan(to_data[2, 0]))

    def test_max_01(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.max(from_data)

        self.assertAlmostEqual(7.0, to_data[0, 0], 0)
        self.assertAlmostEqual(14.0, to_data[1, 0], 0)

    def test_max_02(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 0).tolist()

        to_data = tc.max(from_data)

        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 7 - to_data[0].round() == 0.0
            )
        )
        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 14 - to_data[1].round() == 0.0
            )
        )

    def test_max_03(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=2
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1).tolist()

        to_data = tc.max(from_data, time_dimension=1)

        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 7 - to_data[:, 0, :].round() == 0.0
            )
        )
        self.assertTrue(
            np.all(
                np.ones((3, 4), dtype="int") * 14 - to_data[:, 1, :].round() == 0.0
            )
        )

    def test_max_04(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = RollingWindowTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            end_date=date(2019, 1, 15),
            window_size=7
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.moveaxis(np.asarray(list(range(1, 15)) * 12).reshape((3, 4, 14)), 2, 1).tolist()

        to_data = tc.max(from_data, time_dimension=1)

        for i in range(tc.to_nelem):
            self.assertTrue(
                np.all(
                    np.ones((3, 4), dtype="int") * (7+i) - to_data[:, i, :].round() == 0.0
                )
            )

    def test_max_05(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=14
        ).build()

        to_axis = WeeklyTimeAxisBuilder(
            start_date=date(2019, 1, 1),
            n_interval=3
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = list(range(1, 15))

        to_data = tc.max(from_data)

        self.assertAlmostEqual(7.0, to_data[0, 0], 0)
        self.assertAlmostEqual(14.0, to_data[1, 0], 0)
        self.assertTrue(np.isnan(to_data[2, 0]))


    @skip
    def test_speed_01(self):
        from_axis = DailyTimeAxisBuilder(
            start_date=date(2000, 1, 1),
            end_date=date(2020, 1, 1)
        ).build()

        to_axis = MonthlyTimeAxisBuilder(
            start_year=2000,
            end_year=2019
        ).build()

        tc = AxisConverter(from_axis=from_axis, to_axis=to_axis)

        from_data = np.random.random((from_axis.nelem, 100, 100))

        import time
        start = time.time()
        for i in range(1):
            tc.min(from_data)
        end = time.time()

        print(f"Took: %f [s]", end -start)

        import time
        start = time.time()
        for i in range(1):
            tc.min(from_data)
        end = time.time()

        print(f"Took: %f [s]", end - start)





from abc import ABCMeta, abstractmethod

from axisutilities import Axis


class AxisBuilder(metaclass=ABCMeta):
    @abstractmethod
    def prebuild_check(self) -> (bool, Exception):
        pass

    @abstractmethod
    def build(self) -> Axis:
        pass



from json import JSONEncoder
from ._version import __version__

from .axisbinding import AxisBinding
from .core import Interval, Axis
from .timeaxisbuilders import IntervalBaseTimeAxisBuilder, FixedIntervalTimeAxisBuilder, \
    DailyTimeAxisBuilder, WeeklyTimeAxisBuilder, TimeAxisBuilderFromDataTicks, \
    RollingWindowTimeAxisBuilder, MonthlyTimeAxisBuilder

from .axisconverter import AxisConverter





from unittest import TestCase

from axisutilities import AxisBinding


class TestTimeAxisBinding(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        import os
        os.environ['TZ'] = 'MST'

    def test_repr_01(self):
        self.assertEqual("beginning", repr(AxisBinding.BEGINNING))

    def test_repr_02(self):
        self.assertEqual("end", repr(AxisBinding.END))

    def test_repr_03(self):
        self.assertEqual("middle", repr(AxisBinding.MIDDLE))

    def test_repr_04(self):
        self.assertEqual("custom_fraction", repr(AxisBinding.CUSTOM_FRACTION))

    def test_valueOf_01(self):
        self.assertEqual(
            AxisBinding.BEGINNING,
            AxisBinding.valueOf("beGiNnIng")
        )

    def test_valueOf_02(self):
        with self.assertRaises(ValueError):
            AxisBinding.valueOf("None Existing")

    def test_valueOf_03(self):
        self.assertEqual(
            AxisBinding.BEGINNING,
            AxisBinding.valueOf(AxisBinding.BEGINNING)
        )

    def test_valueOf_04(self):
        with self.assertRaises(TypeError):
            AxisBinding.valueOf({})

    def test_valueOf_05(self):
        self.assertEqual(AxisBinding.BEGINNING, AxisBinding.valueOf(0))

    def test_valueOf_06(self):
        self.assertEqual(AxisBinding.END, AxisBinding.valueOf(1.0))

    def test_valueOf_07(self):
        self.assertEqual(AxisBinding.MIDDLE, AxisBinding.valueOf(0.5))

    def test_valueOf_08(self):
        self.assertEqual(AxisBinding.CUSTOM_FRACTION, AxisBinding.valueOf(0.2))

    def test_valueOf_09(self):
        with self.assertRaises(ValueError):
            AxisBinding.valueOf(10)

    def test_fraction_01(self):
        self.assertEqual(0.0, AxisBinding.BEGINNING.fraction())

    def test_fraction_02(self):
        self.assertEqual(1.0, AxisBinding.END.fraction())

    def test_fraction_03(self):
        self.assertEqual(0.5, AxisBinding.MIDDLE.fraction())

    def test_fraction_04(self):
        self.assertEqual(None, AxisBinding.CUSTOM_FRACTION.fraction())









from __future__ import annotations

from typing import Iterable, Callable

import numpy as np
from numba import prange
from scipy.sparse import csr_matrix

from axisutilities import Axis


class AxisConverter:
    def __init__(self, **kwargs) -> None:
        if ("from_axis" in kwargs) and ("to_axis" in kwargs):
            from_ta = kwargs["from_axis"]
            to_ta = kwargs["to_axis"]
            if not (isinstance(from_ta, Axis) and isinstance(to_ta, Axis)):
                raise TypeError("provided from/to_axis must be of type TimeAxis.")

            self._m = to_ta.nelem
            self._n = from_ta.nelem
            self._weight_matrix = self._get_coverage_csr_matrix(from_ta, to_ta)
            self._from_ta = from_ta
            self._to_ta = to_ta
        else:
            raise ValueError("Not enough information is provided to construct the TimeAxisConverter.")

    @property
    def from_nelem(self):
        return self._n

    @from_nelem.setter
    def from_nelem(self, v):
        pass

    @property
    def to_nelem(self):
        return self._m

    @to_nelem.setter
    def to_nelem(self, v):
        pass

    @property
    def weights(self) -> csr_matrix:
        return self._weight_matrix.copy()

    @weights.setter
    def weights(self, v):
        pass

    @property
    def from_axis(self):
        return self._from_ta

    @from_axis.setter
    def from_axis(self, v):
        pass

    @property
    def to_axis(self):
        return self._to_ta

    @to_axis.setter
    def to_axis(self, v):
        pass

    def _prep_input_data(self, in_data: Iterable, time_dimension) -> (np.ndarray, tuple):
        if not isinstance(in_data, Iterable):
            raise TypeError("input data should be an Iterable that can be casted to numpy.ndarray.")

        in_data_copy = in_data
        if not isinstance(in_data_copy, np.ndarray):
            in_data_copy = np.asarray(in_data_copy, dtype="float64")

        if in_data_copy.ndim == 1:
            in_data_copy = in_data_copy.reshape((-1, 1))

        n = self.from_nelem

        if in_data_copy.shape[time_dimension] != n:
            raise ValueError("The time dimension does not matches to that of the provided time converter.")

        if time_dimension != 0:
            in_data_copy = np.moveaxis(in_data_copy, time_dimension, 0)

        trailing_shape = in_data_copy.shape[1:]
        in_data_copy = in_data_copy.reshape((n, -1))

        return in_data_copy, trailing_shape

    @staticmethod
    def _prep_output_data( out_data: np.ndarray, time_dimension, trailing_shape: tuple):
        return np.moveaxis(out_data.reshape((out_data.shape[0], *trailing_shape)), 0, time_dimension)

    def average(self, from_data: Iterable, time_dimension=0) -> np.ndarray:
        from_data_copy, trailing_shape = self._prep_input_data(from_data, time_dimension)

        nan_mask = np.isnan(from_data_copy)
        non_nan_mask = np.ones(from_data_copy.shape, dtype=np.int8)
        non_nan_mask[nan_mask] = 0
        from_data_copy[nan_mask] = 0

        inverse_sum_effective_weights = np.reciprocal(self._weight_matrix * non_nan_mask)

        return self._prep_output_data(
            np.multiply(self._weight_matrix * from_data_copy, inverse_sum_effective_weights),
            time_dimension,
            trailing_shape
        )

    def apply_function(self, data: Iterable, func2apply: Callable, time_dimension=0):
        data_copy, trailing_shape = self._prep_input_data(data, time_dimension)

        if isinstance(func2apply, Callable):
            import warnings
            warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
            output = _apply_function_core(
                self.to_nelem,
                self._weight_matrix,
                data_copy,
                func2apply.__call__
            )
        else:
            raise TypeError("func2apply must be a callable object that performs the calculation on axis=0.")

        return self._prep_output_data(
            output,
            time_dimension,
            trailing_shape
        )

    def min(self, data, time_dimension=0):
        return self.apply_function(
            data,
            np.nanmin,
            time_dimension
        )

    def max(self, data, time_dimension=0):
        return self.apply_function(
            data,
            np.nanmax,
            time_dimension
        )



    @staticmethod
    def _get_coverage_csr_matrix(from_ta: Axis, to_ta: Axis) -> csr_matrix:
        row_idx, col_idx, weights = AxisConverter._get_coverage(
            from_ta.lower_bound, from_ta.upper_bound,
            to_ta.lower_bound, to_ta.upper_bound
        )
        m = to_ta.nelem
        n = from_ta.nelem
        weights = csr_matrix((weights, (row_idx, col_idx)), shape=(m, n)).tolil()
        # with np.errstate(divide='ignore'):
        #     row_sum_reciprocal = np.reciprocal(np.asarray(weights.sum(axis=1)).flatten())
        # mask = np.isinf(row_sum_reciprocal)
        # row_sum_reciprocal[mask] = 0.0
        # inverse_row_sum = spdiags(row_sum_reciprocal, 0, m, m)
        #
        # normalized_weights = (inverse_row_sum * weights).tolil()
        # normalized_weights[mask, 0] = np.nan
        #
        # return normalized_weights.tocsr()

        mask = np.asarray(weights.sum(axis=1)).flatten() == 0
        weights[mask, 0] = np.nan
        return weights.tocsr()

    @staticmethod
    def _get_coverage(
            from_lower_bound: np.ndarray,
            from_upper_bound: np.ndarray,
            to_lower_bound: np.ndarray,
            to_upper_bound: np.ndarray):
        m = to_lower_bound.size
        n = from_lower_bound.size

        # basic sanity checks:
        if (to_lower_bound.ndim != 2) or (to_lower_bound.shape[0] != 1):
            raise ValueError(f"to_lower_bound must be of shape (1,m), it's current shape is: {to_lower_bound.shape}.")

        if to_lower_bound.shape != to_upper_bound.shape:
            raise ValueError("to_lower_bound/upper_bound must have the same shape.")

        if (from_lower_bound.ndim != 2) or (from_lower_bound.shape[0] != 1):
            raise ValueError("from_lower_bound must be of shape (1,n).")

        if from_lower_bound.shape != from_upper_bound.shape:
            raise ValueError("from_lower_bound/upper_bound must have the same shape.")

        # if np.any(from_lower_bound[0, :-1] > from_lower_bound[0, 1:]):
        #     raise ValueError("from_lower_bound must be monotonically increasing.")

        # TODO: turn this into cython so that is faster and/or use some sort of data structure to
        #       reduce its time-complexity from O(mn)
        row_idx = []
        col_idx = []
        weights = []
        for r in range(m):
            toLB = to_lower_bound[0, r]
            toUB = to_upper_bound[0, r]
            for c in range(n):
                fromLB = from_lower_bound[0, c]
                fromUB = from_upper_bound[0, c]
                fromLength = fromUB - fromLB

                if (fromUB <= toLB) or (fromLB >= toUB):  # No coverage
                    continue
                elif (fromLB <= toLB) and (fromUB >= toLB) and (fromUB <= toUB):
                    row_idx.append(r)
                    col_idx.append(c)
                    weights.append((fromUB - toLB) / fromLength)
                elif (fromLB >= toLB) and (fromLB < toUB) and (fromUB >= toUB):
                    row_idx.append(r)
                    col_idx.append(c)
                    weights.append((toUB - fromLB) / fromLength)
                elif (fromLB >= toLB) and (fromUB <= toUB):
                    row_idx.append(r)
                    col_idx.append(c)
                    weights.append(1.0)
                elif (fromLB <= toLB) and (fromUB >= toUB):
                    row_idx.append(r)
                    col_idx.append(c)
                    weights.append((toUB - toLB) / fromLength)

        return row_idx, col_idx, weights


# @jit(parallel=True, forceobj=True, cache=True)
# @autojit
def _apply_function_core(n: int, _weight_matrix: csr_matrix, data_copy: np.ndarray, func: Callable) -> np.ndarray:
    output = np.full((n, data_copy.shape[1]), np.nan)
    for r in prange(n):
        start = _weight_matrix.indptr[r]
        end = _weight_matrix.indptr[r + 1]
        if not (np.isnan(_weight_matrix[r, 0]) and ((end - start) == 1)):
            row_mask = _weight_matrix.indices[start:end]
            output[r, :] = func(data_copy[row_mask, :], axis=0)
    return output

























from datetime import datetime

_today_datetime = datetime.now()

__version__ = f"{_today_datetime.year % 100}." \
              f"{_today_datetime.month}." \
              f"{_today_datetime.day}." \
              f"{_today_datetime.hour * 3600 + _today_datetime.minute * 60 + _today_datetime.second}"

del _today_datetime


from __future__ import annotations
import json

from datetime import datetime
from typing import Iterable, Dict

import numpy as np

from axisutilities import AxisBinding
from axisutilities.constants import SECONDS_TO_MICROSECONDS_FACTOR


class Interval:
    def __init__(self, lower_bound: int, upper_bound: int, data_tick: int):
        if not isinstance(lower_bound, int) or \
           not isinstance(upper_bound, int) or \
           not isinstance(data_tick, int):
            raise TypeError("All inputs must be of type int.")

        if (lower_bound > upper_bound) or \
           (data_tick > upper_bound) or \
           (data_tick < lower_bound):
            raise ValueError("Invalid inputs. Check lower_/upper_bound and data_tick")

        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._data_tick = data_tick
        self._fraction = (data_tick - lower_bound) / (upper_bound - lower_bound)

        self._binding = AxisBinding.valueOf(self._fraction)

    def asDict(self):
        return {
            "lower_bound": str(Interval.timestamp_to_datetime(self._lower_bound)),
            "upper_bound": str(Interval.timestamp_to_datetime(self._upper_bound)),
            "data_tick": str(Interval.timestamp_to_datetime(self._data_tick)),
            "fraction": self._fraction,
            "binding": str(self._binding)
        }

    def asJson(self):
        return json.dumps(self.asDict())

    def __repr__(self):
        summary = ["<timeaxis.TimeInterval>\n"]
        output = self.asDict()
        for key, value in output.items():
            if key in {"binding", "fraction"}:
                v = f"  > {key}:\n\t{value}\n"
            else:
                v = f"  > {key}:\n\t[{value[0]} ... {value[-1]}]\n"
            summary.append(v)
        return "\n".join(summary)

    @staticmethod
    def timestamp_to_datetime(ts: int) -> datetime:
        return datetime.fromtimestamp(ts/SECONDS_TO_MICROSECONDS_FACTOR)

    @property
    def lower_bound(self) -> int:
        return self._lower_bound

    @lower_bound.setter
    def lower_bound(self, v):
        pass

    @property
    def upper_bound(self) -> int:
        return self._upper_bound

    @upper_bound.setter
    def upper_bound(self, v):
        pass

    @property
    def data_tick(self) -> int:
        return self._data_tick

    @data_tick.setter
    def data_tick(self, v):
        pass


class Axis:
    def __init__(self,  lower_bound: Iterable[float],
                        upper_bound: Iterable[float],
                        **kwargs):
        """
        Initializes a ``TimeAxis`` object given the ``lower_bound``, ``upper_bound``, and either ``fraction`` or
        ``data_ticks``.

        :param lower_bound: The lower bound time stamps as a ``long`` value (or something that could be casted to
                            ``long``). The units must be in microseconds. The ``lower_bound`` values must be
                            monotonically increasing.
        :param upper_bound: The upper bound time stamps as a ``long`` value (or something that could be casted to
                            ``long``). The units must be in microseconds. The ``upper_bound`` does NOT need to be
                            monotonically increasing; however, they should not be smaller than their counter-part
                            ``lower_bound``.
        :param kwargs: One of the following keys (just one, not more) must be provided:
            - ``fraction``:
            - ``data_ticks``:
            - ``binding``:
        """
        # making sure at there is only one of the "fraction", "data_ticks", or "binding" are provided.
        if sum(list(map(lambda e: 1 if e in kwargs else 0, ['fraction', 'data_ticks', 'binding']))) != 1:
            raise ValueError("You must provide exactly just one of the 'fraction', 'data_ticks', or 'binding'.")

        self._bounds = Axis._create_bounds(lower_bound, upper_bound)
        self._nelem = self._bounds.shape[1]

        if "fraction" in kwargs:
            self._fraction = np.asarray(kwargs["fraction"], dtype="float64").reshape((1, -1))
            Axis._fraction_sanity_check(self._fraction)
            if (self._fraction.size != 1) and (self._fraction.size != self._nelem):
                raise ValueError("Fraction must be either a single number, or as many as there "
                                 "upper/lower bound values.")

            self._data_ticks = Axis._calculate_data_ticks_fromFraction(
                self._bounds[0, :],
                self._bounds[1, :],
                self._fraction
            )
            self._binding = Axis._get_binding(self._fraction)

        if "data_ticks" in kwargs:
            self._data_ticks = np.asarray(kwargs["data_ticks"], dtype="int64").reshape((1, -1))
            Axis._data_ticks_sanity_check(self._data_ticks)
            if self._data_ticks.size != self._nelem:
                raise ValueError("Data Ticks must have as many elements as there are lower/upper bound values.")

            self._fraction = Axis._calculate_fraction_from_data_ticks(
                self._bounds[0, :],
                self._bounds[1, :],
                self._data_ticks
            )
            self._binding = Axis._get_binding(self._fraction)

        if "binding" in kwargs:
            self._binding = AxisBinding.valueOf(kwargs["binding"])
            self._fraction = np.asarray(self._binding.fraction(), dtype="float64").reshape((1, -1))
            self._data_ticks = Axis._calculate_data_ticks_fromFraction(
                self._bounds[0, :],
                self._bounds[1, :],
                self._fraction
            )

    def asDict(self) -> Dict:
        out = {
            "nelem": self.nelem,
            "lower_bound": self._bounds[0, :].tolist(),
            "upper_bound": self._bounds[1, :].tolist(),
            "data_ticks": self._data_ticks.tolist()[0],
            "fraction": self._fraction.tolist()[0],
            "binding": str(self._binding)
        }

        return out

    def asJson(self) -> str:
        return json.dumps(self.asDict())

    def __repr__(self):
        summary = ["<timeaxis.TimeAxis>\n"]
        output = self.asDict()
        for key, value in output.items():
            if key in {"binding", "nelem"}:
                v = f"  > {key}:\n\t{value}\n"
            else:
                v = f"  > {key}:\n\t[{value[0]} ... {value[-1]}]\n"
            summary.append(v)
        return "\n".join(summary)

    def __getitem__(self, item: int) -> Interval:
        if isinstance(item, int):
            if (item >= self.nelem) or (item < -self.nelem):
                raise IndexError("Index Out of range")
            return Interval(
                int(self._bounds[0, item]),
                int(self._bounds[1, item]),
                int(self._data_ticks[0, item])
            )
        else:
            raise TypeError("item must be an integer")

    @property
    def nelem(self) -> int:
        return self._nelem

    @nelem.setter
    def nelem(self, v) -> None:
        pass

    @property
    def lower_bound(self) -> np.ndarray:
        return self._bounds[0, :].copy().reshape((1, -1))

    @lower_bound.setter
    def lower_bound(self, v) -> None:
        pass

    @property
    def upper_bound(self) -> np.ndarray:
        return self._bounds[1, :].copy().reshape((1, -1))

    @upper_bound.setter
    def upper_bound(self, v) -> None:
        pass

    @property
    def fraction(self) -> np.ndarray:
        return self._fraction.copy()

    @fraction.setter
    def fraction(self, v) -> None:
        pass

    @property
    def data_ticks(self) -> np.ndarray:
        return self._data_ticks.copy()

    @data_ticks.setter
    def data_ticks(self, v) -> None:
        pass

    @staticmethod
    def _bounds_sanity_check(bounds: np.ndarray) -> (bool, Exception):
        # Making sure lower_i <= upper_i
        if bounds.shape[1] > 1:
            if np.any(bounds[:, 0] > bounds[:, 1]):
                raise ValueError("all lower bounds must be smaller than their counter-part upper bounds")

            # making sure lower_i < lower_{i+1}
            if np.any(bounds[0, :-1] >= bounds[0, 1:]):
                raise ValueError('lower bound values must be monotonically increasing.')

        # All good
        return True

    @staticmethod
    def _create_bounds(lower_bound: Iterable[float], upper_bound: Iterable[float]) -> np.ndarray:
        _bounds = np.asarray([list(lower_bound), list(upper_bound)], dtype="int64")
        Axis._bounds_sanity_check(_bounds)
        return _bounds

    @staticmethod
    def _data_ticks_sanity_check(data_tick: np.ndarray) -> (bool, Exception):
        # making sure data_tick_i < data_tick_{i+1}
        if np.any(data_tick[0, :-1] >= data_tick[0, 1:]):
            raise ValueError("data_tick must be monotonically increasing.")

        return True

    @staticmethod
    def _calculate_data_ticks_fromFraction(
            lower_bound: np.ndarray,
            upper_bound: np.ndarray,
            f: np.ndarray) -> np.ndarray:

        if any(f < 0) or any(f > 1):
            raise ValueError("all values of fraction must be between 0 and 1")
        data_tick = ((1.0 - f) * lower_bound + f * upper_bound).astype(np.int64)
        Axis._data_ticks_sanity_check(data_tick)
        return data_tick

    @ staticmethod
    def _fraction_sanity_check(fraction: np.ndarray) -> (bool, Exception):
        if np.any(fraction < 0):
            raise ValueError("all the data ticks values must be between their lower/upper bound.")

        return True

    @staticmethod
    def _calculate_fraction_from_data_ticks(
            lower_bound: np.ndarray,
            upper_bound: np.ndarray,
            data_ticks: np.ndarray) -> np.ndarray:

        fraction = ((data_ticks - lower_bound) /
                    (upper_bound - lower_bound)).astype(np.float64)

        Axis._fraction_sanity_check(fraction)

        return fraction

    @staticmethod
    def _get_binding(fraction: np.ndarray) -> AxisBinding:
        if np.all(fraction == 0.0):
            return AxisBinding.BEGINNING
        elif np.all(fraction == 1.0):
            return AxisBinding.END
        elif np.all(fraction == 0.5):
            return AxisBinding.MIDDLE
        elif np.all(0.0 < fraction < 1.0):
            return AxisBinding.CUSTOM_FRACTION
        else:
            raise ValueError("fraction must be a number between 0.0 and 1.0")

    # @staticmethod
    # def builder(name: str) -> TimeAxisBuilder:
    #     if isinstance(name, str):
    #         name_lower = name.lower()
    #         if name_lower == "fixed_interval":
    #             return FixedIntervalTimeAxisBuilder()
    #
    #         if name_lower == "daily_time_axis":
    #             return DailyTimeAxisBuilder()
    #
    #         if name_lower == "weekly_time_axis":
    #             return WeeklyTimeAxisBuilder()
    #
    #         if name_lower == "rolling_window":
    #             return RollingWindowTimeAxisBuilder()
    #
    #         raise ValueError(f"Could not find a TimaAxisBUilder associated to {name}.")
    #     else:
    #         raise TypeError("name must be string")









from unittest import TestCase

from axisutilities import Interval


class TestTimeInterval(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls._sample_data = []

        # _sample_data[0]: everything is ok:
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546369200000000])

        # _sample_data[1]: upper_bound is smaller than lower_bound
        cls._sample_data.append([1546412400000000, 1546326000000000, 1546369200000000])

        # _sample_data[2]: data_tick is less than lower_bound:
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546325000000000])

        # _sample_data[3]: data_tick is larger than the upper_bound:
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546412500000000])

        # _sample_data[4]: fraction is 0.25
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546347600000000])

        # _sample_data[5]: fraction is 0.0
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546326000000000])

        # _sample_data[6]: fraction is 1.0
        cls._sample_data.append([1546326000000000, 1546412400000000, 1546412400000000])

        import os
        os.environ['TZ'] = "MST"

    def test_creation_00(self):
        lower_bound = self._sample_data[0][0]
        upper_bound = self._sample_data[0][1]
        data_tick = self._sample_data[0][2]

        ti = Interval(lower_bound, upper_bound, data_tick)
        print("Sample TimeInterval: ", ti.asJson())
        expected = '{"lower_bound": "2019-01-01 00:00:00", ' \
                    '"upper_bound": "2019-01-02 00:00:00", ' \
                    '"data_tick": "2019-01-01 12:00:00", ' \
                    '"fraction": 0.5, ' \
                    '"binding": "middle"}'
        self.assertEqual(expected, ti.asJson())

    def test_creation_01(self):
        lower_bound = self._sample_data[1][0]
        upper_bound = self._sample_data[1][1]
        data_tick = self._sample_data[1][2]

        with self.assertRaises(ValueError):
            Interval(lower_bound, upper_bound, data_tick)

    def test_creation_02(self):
        lower_bound = self._sample_data[2][0]
        upper_bound = self._sample_data[2][1]
        data_tick = self._sample_data[2][2]

        with self.assertRaises(ValueError):
            Interval(lower_bound, upper_bound, data_tick)

    def test_creation_03(self):
        lower_bound = self._sample_data[3][0]
        upper_bound = self._sample_data[3][1]
        data_tick = self._sample_data[3][2]

        with self.assertRaises(ValueError):
            Interval(lower_bound, upper_bound, data_tick)

    def test_creation_04(self):
        lower_bound = self._sample_data[4][0]
        upper_bound = self._sample_data[4][1]
        data_tick = self._sample_data[4][2]

        ti = Interval(lower_bound, upper_bound, data_tick)
        print("Sample TimeInterval: ", ti.asJson())
        expected = '{"lower_bound": "2019-01-01 00:00:00", ' \
                    '"upper_bound": "2019-01-02 00:00:00", ' \
                    '"data_tick": "2019-01-01 06:00:00", ' \
                    '"fraction": 0.25, ' \
                    '"binding": "custom_fraction"}'
        self.assertEqual(expected, ti.asJson())

    def test_creation_05(self):
        lower_bound = self._sample_data[5][0]
        upper_bound = self._sample_data[5][1]
        data_tick = self._sample_data[5][2]

        ti = Interval(lower_bound, upper_bound, data_tick)
        print("Sample TimeInterval: ", ti.asJson())
        expected = '{"lower_bound": "2019-01-01 00:00:00", ' \
                    '"upper_bound": "2019-01-02 00:00:00", ' \
                    '"data_tick": "2019-01-01 00:00:00", ' \
                    '"fraction": 0.0, ' \
                    '"binding": "beginning"}'
        self.assertEqual(expected, ti.asJson())

    def test_creation_06(self):
        lower_bound = self._sample_data[6][0]
        upper_bound = self._sample_data[6][1]
        data_tick = self._sample_data[6][2]

        ti = Interval(lower_bound, upper_bound, data_tick)
        print("Sample TimeInterval: ", ti.asJson())
        expected = '{"lower_bound": "2019-01-01 00:00:00", ' \
                    '"upper_bound": "2019-01-02 00:00:00", ' \
                    '"data_tick": "2019-01-02 00:00:00", ' \
                    '"fraction": 1.0, ' \
                    '"binding": "end"}'
        self.assertEqual(expected, ti.asJson())



import setuptools

from axisutilities import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', "r") as fh:
    requirements = fh.read().split("\n")

setuptools.setup(
    name="AxisUtilities",
    version=__version__,
    author="Mohammad Abouali",
    author_email="maboualidev@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/coderepocenter/AxisUtilities.git",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)


from __future__ import annotations
from abc import ABCMeta, ABC, abstractmethod
from calendar import monthrange
from datetime import datetime, date, timedelta
from typing import Iterable

import numpy as np

from axisutilities import Axis
from axisutilities.axisbuilder import AxisBuilder
from axisutilities.constants import SECONDS_TO_MICROSECONDS_FACTOR


class TimeAxisBuilder(AxisBuilder, ABC, metaclass=ABCMeta):
    @staticmethod
    def datetime_to_timestamp(data_ticks: (datetime, date, str, Iterable), **kwrargs) -> np.ndarray:
        if isinstance(data_ticks, date):
            return np.asarray(
                [datetime.strptime(data_ticks.strftime("%c"), "%c").timestamp() * SECONDS_TO_MICROSECONDS_FACTOR],
                dtype="int64"
            )
        elif isinstance(data_ticks, datetime):
            return np.asarray(
                [data_ticks.timestamp() * SECONDS_TO_MICROSECONDS_FACTOR],
                dtype="int64"
            )
        elif isinstance(data_ticks, str):
            raise NotImplemented("")
        elif isinstance(data_ticks, Iterable):
            return np.asarray(
                list(
                    map(lambda e: TimeAxisBuilder.datetime_to_timestamp(e), data_ticks)
                ),
                dtype="int64"
            ).flatten().reshape((1, -1))
        else:
            raise TypeError("data_ticks must be either a single value of type date or datetime, "
                            "or and iterable where all of its elements are of type date or datetime.")


class IntervalBaseTimeAxisBuilder(TimeAxisBuilder):
    _key_properties = ['_start', '_end', '_interval']

    def __init__(self, **kwargs):
        self.set_start(kwargs.get("start", None))
        self.set_end(kwargs.get("end", None))
        self.set_interval(kwargs.get("interval", None))
        self.set_fraction(kwargs.get("fraction", 0.5))

    def set_start(self, start: int) -> FixedIntervalTimeAxisBuilder:
        if (start is None) or isinstance(start, int):
            self._start = start
        else:
            raise TypeError("start must be an int or None.")

        return self

    def set_end(self, end: int) -> FixedIntervalTimeAxisBuilder:
        if (end is None) or isinstance(end, int):
            self._end = end
        else:
            raise TypeError("end must be an int or None.")

        return self

    def set_interval(self, interval: [int, Iterable]) -> FixedIntervalTimeAxisBuilder:
        if isinstance(interval, int):
            self._interval = interval
        if isinstance(interval, Iterable):
            self._interval = np.asarray(list(interval), dtype=np.int64).reshape((1, -1))
        elif interval is None:
            self._interval = None
        else:
            raise TypeError("interval must be an int, an Iterable, or None.")

        return self

    def set_fraction(self, fraction: float) -> FixedIntervalTimeAxisBuilder:
        if isinstance(fraction, float):
            self._fraction = fraction
        elif fraction is None:
            self._fraction = 0.5
        else:
            raise TypeError("start must be a float or None.")

        if (self._fraction < 0) or (self._fraction > 1):
            raise ValueError("Fraction must be between 0 and 1.")

        return self

    def prebuild_check(self) -> (bool, Exception):
        if (self._start is None) or (self._end is None) or (self._interval is None):
            raise ValueError("Not yet Ready to build the time axis.")

        if self._start > self._end:
            raise ValueError("Start must be smaller than  end")

        if (self._fraction is None) or (self._fraction < 0) or (self._fraction > 1):
            raise ValueError("some how fraction ended up to be None or out of bounds. "
                             f"Current Fraction Value: {self._fraction}")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            if isinstance(self._interval, int):
                lower_bound = np.arange(self._start, self._end, self._interval)
                upper_bound = np.arange(self._start + self._interval, self._end + 1, self._interval, dtype="int64")
            elif isinstance(self._interval, np.ndarray):
                interval_cumsum = np.concatenate(
                    (np.zeros((1, 1)),self._interval.cumsum().reshape((1, -1))),
                    axis=1
                ).astype(dtype="int64")
                lower_bound = self._start + interval_cumsum[0,:-1]
                upper_bound = self._start + interval_cumsum[0,1:]
            else:
                raise TypeError("Somehow interval ended up to be of a type other than an int or numpy.ndarray "
                                "(Iterable)")

            data_ticks = (1 - self._fraction) * lower_bound + self._fraction * upper_bound
            return Axis(lower_bound, upper_bound, data_ticks=data_ticks)


class FixedIntervalTimeAxisBuilder(IntervalBaseTimeAxisBuilder):
    _key_properties = ['_start', '_end', '_interval', '_n_interval']

    def __init__(self, **kwargs):
        self.set_n_interval(kwargs.get("n_interval", None))
        super().__init__(**kwargs)

    def set_interval(self, interval: int) -> FixedIntervalTimeAxisBuilder:
        if isinstance(interval, int):
            self._interval = interval
        elif interval is None:
            self._interval = None
        else:
            raise TypeError("interval must be an int or None.")

        return self

    def set_n_interval(self, n_interval: int) -> FixedIntervalTimeAxisBuilder:
        if isinstance(n_interval, int):
            self._n_interval = n_interval
        elif n_interval is None:
            self._n_interval = None
        else:
            raise TypeError("n_interval must be an int or None.")

        return self

    def prebuild_check(self) -> (bool, Exception):
        self._mask = 0
        self._n_available_keys = 0
        for idx in range(len(self._key_properties)):
            self._mask += 2 ** idx if self.__getattribute__(self._key_properties[idx]) is not None else 0
            self._n_available_keys += 1 if self.__getattribute__(self._key_properties[idx]) is not None else 0

        if self._n_available_keys != 3:
            raise ValueError(f"Only three out of the four {self._key_properties} "
                             f"could/should be provided. "
                             f"Currently {self._n_available_keys} are provided.")

        if self._mask not in {7, 11, 13, 14}:
            raise ValueError("wrong combination of inputs are provided.")

        if (self._start is not None) and (self._end is not None) and (self._start >= self._end):
            raise ValueError("start must be less than end.")

        if (self._interval is not None) and (self._interval<= 0.0):
            raise ValueError("interval must be a positive number.")

        if (self._n_interval is not None) and (self._n_interval < 1):
            raise ValueError("n_interval must be larger than 1.")

        if (self._fraction is None) or (self._fraction < 0) or (self._fraction > 1):
            raise ValueError("some how fraction ended up to be None or out of bounds. "
                             f"Current Fraction Value: {self._fraction}")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            if self._mask == 7:
                # this means start, end, and interval are provided
                self._n_interval = int((self._end - self._start)/self._interval)

                if self._n_interval < 1:
                    raise ValueError("provided input leaded to wrong n_interval.")

                if (self._start + self._n_interval * self._interval) != self._end:
                    raise ValueError("provided interval does not divide the start to end interval properly.")
            if self._mask == 11:
                # this means start, end, and n_interval are provided
                self._interval = int((self._end - self._start)/self._n_interval)

            if self._mask == 13:
                # this means start, interval, and n_interval are provided
                self._end = self._start + (self._n_interval + 1) * self._interval

            if self._mask == 14:
                # this means end, interval, and n_interval are provided
                self._start = self._end - (self._n_interval + 1) * self._interval

            lower_bound = self._start + np.arange(self._n_interval, dtype="int64") * self._interval
            upper_bound = lower_bound + self._interval
            if upper_bound[-1] != self._end:
                raise ValueError(f"last element of upper_bound (i.e. {upper_bound[-1]}) is not the same "
                                 f"as provided end (i.e. {self._end}).")

            data_ticks = (1 - self._fraction) * lower_bound + self._fraction * upper_bound
            return Axis(lower_bound, upper_bound, data_ticks=data_ticks)


class BaseCommonKnownIntervals(TimeAxisBuilder, metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def get_dt() -> int:
        pass

    def __init__(self, **kwargs):
        self.set_start_date(kwargs.get("start_date", None))
        self.set_end_date(kwargs.get("end_date", None))
        self.set_n_interval(kwargs.get("n_interval", None))

    def set_start_date(self, start_date: date) -> BaseCommonKnownIntervals:
        if (start_date is None) or isinstance(start_date, date):
            self._start_date = start_date
        else:
            raise TypeError("start_date must be of type date.")

        return self

    def set_end_date(self, end_date: date) -> BaseCommonKnownIntervals:
        if (end_date is None) or isinstance(end_date, date):
            self._end_date = end_date
        else:
            raise TypeError("end_date must be of type date.")

        return self

    def set_n_interval(self, n_interval: int) -> BaseCommonKnownIntervals:
        self._n_interval = None if n_interval is None else int(n_interval)
        return self

    def prebuild_check(self) -> (bool, Exception):
        if sum(list(map(
                lambda e: 1 if self.__getattribute__(e) is not None else 0,
                ["_start_date", "_end_date", "_n_interval"]))) != 2:
            raise ValueError('Only two out of the "_start_date", "_end_date", or "_n_interval" could be provided.')

        if (self._start_date is not None) and \
           (self._end_date is not None) and \
           (self._start_date > self._end_date):
            raise ValueError("start_date cannot be larger than end_date.")

        if (self._n_interval is not None) and (self._n_interval < 1):
            raise ValueError("n_interval must be at least 1.")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            # # Older implementation
            # dt = np.int64(timedelta(days=1).total_seconds() * SECONDS_TO_MICROSECONDS_FACTOR)
            # nDays = (self._end_date - self._start_date).days + 1
            # lower_bound = np.arange(nDays, dtype="int64") * dt + TimeAxisBuilder.datetime_to_timestamp(self._start_date)
            # upper_bound = lower_bound + dt
            # data_ticks = lower_bound + np.int64(timedelta(hours=12).total_seconds() * SECONDS_TO_MICROSECONDS_FACTOR)

            # return TimeAxis(lower_bound, upper_bound, data_ticks=data_ticks)

            if (self._start_date is not None) and (self._end_date is not None):
                start = int(TimeAxisBuilder.datetime_to_timestamp(self._start_date))
                dt = self.get_dt()
                end = int(TimeAxisBuilder.datetime_to_timestamp(self._end_date))
                return FixedIntervalTimeAxisBuilder(start=start, end=end, interval=dt).build()

            if (self._start_date is not None) and (self._n_interval is not None):
                start = int(TimeAxisBuilder.datetime_to_timestamp(self._start_date))
                dt = self.get_dt()
                end = start + self._n_interval * dt
                return FixedIntervalTimeAxisBuilder(start=start, end=end, interval=dt).build()

            if (self._end_date is not None) and (self._n_interval is not None):
                end = int(TimeAxisBuilder.datetime_to_timestamp(self._end_date))
                dt = self.get_dt()
                start = end - self._n_interval * dt
                return FixedIntervalTimeAxisBuilder(start=start, end=end, interval=dt).build()


class DailyTimeAxisBuilder(BaseCommonKnownIntervals):
    @staticmethod
    def get_dt() -> int:
        return int(timedelta(days=1).total_seconds() * SECONDS_TO_MICROSECONDS_FACTOR)


class WeeklyTimeAxisBuilder(BaseCommonKnownIntervals):
    @staticmethod
    def get_dt() -> int:
        return int(timedelta(days=7).total_seconds() * SECONDS_TO_MICROSECONDS_FACTOR)


class TimeAxisBuilderFromDataTicks(TimeAxisBuilder):
    _acceptable_boundary_types = {
        "centered"
    }

    def __init__(self, data_ticks=None, boundary_type="centered", **kwargs):
        if data_ticks is None:
            self._data_ticks = None
        else:
            self.set_data_ticks(data_ticks)

        self._boundary_type = "centered"
        self.set_boundary_type(boundary_type)

    def set_data_ticks(self, data_ticks: Iterable) -> TimeAxisBuilderFromDataTicks:
        self._data_ticks = TimeAxisBuilder.datetime_to_timestamp(data_ticks)
        return self

    def set_boundary_type(self, boundary_type) -> TimeAxisBuilderFromDataTicks:
        if isinstance(boundary_type, str):
            boundary_type_lower = boundary_type.lower()
            if boundary_type_lower in self._acceptable_boundary_types:
                self._boundary_type = boundary_type_lower
            else:
                raise ValueError(f"Unrecognized boundary type. Currently acceptable values are: "
                                 f"[{', '.join(self._acceptable_boundary_types)}].")
        else:
            raise TypeError(f"boundary_type must be a string set to one of the "
                            f"following values: {str(self._acceptable_boundary_types)}")

        return self

    def prebuild_check(self) -> (bool, Exception):
        if self._data_ticks is None:
            raise ValueError("data_ticks are not set yet.")

        if self._boundary_type is None:
            raise ValueError("Boundary Type is not provided.")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            lower_bound, data_tickes, upper_bound = TimeAxisBuilderFromDataTicks._calculate_bounds(
                data_ticks=self._data_ticks,
                boundary_type=self._boundary_type
            )

            return Axis(
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                data_ticks=data_tickes
            )

    @staticmethod
    def _calculate_bounds(
            data_ticks: np.ndarray,
            boundary_type: str = "centered") -> tuple[np.ndarray, np.ndarray, np.ndarray]:

        if not isinstance(data_ticks, np.ndarray):
            raise TypeError("This method only accepts numpy.ndarry.")

        if data_ticks.ndim == 1:
            data_ticks = data_ticks.reshape((-1, ))

        if (data_ticks.ndim == 2) and ((data_ticks.shape[0] != 1) or ((data_ticks.shape[1] != 1))):
            data_ticks = data_ticks.reshape((-1, ))

        if (data_ticks.ndim != 1):
            raise ValueError("data_ticks must be a row/column of numbers.")

        if data_ticks.dtype != 'int64':
            data_ticks = data_ticks.astype(np.int64)

        boundary_type = boundary_type.lower()
        if boundary_type == "centered":
            avg = (0.5 * (data_ticks[:-1] + data_ticks[1:])).astype(np.int64)

            n = data_ticks.size

            lower_boundary: np.ndarray = np.ndarray((n, ), dtype=np.int64)
            lower_boundary[0] = 2 * data_ticks[0] - avg[0]
            lower_boundary[1:] = avg

            upper_boundary: np.ndarray = np.ndarray((n, ), dtype=np.int64)
            upper_boundary[-1] = 2 * data_ticks[-1] - avg[-1]
            upper_boundary[:-1] = avg

            return lower_boundary, data_ticks, upper_boundary
        else:
            raise ValueError("Unrecognized boundary type.")


class RollingWindowTimeAxisBuilder(TimeAxisBuilder):
    def __init__(self, **kwargs):
        self.set_start_date(kwargs.get("start_date", None))
        self.set_end_date(kwargs.get("end_date", None))
        self.set_n_window(kwargs.get("n_window", None))
        self.set_window_size(kwargs.get("window_size", None))
        self.set_base_dt(kwargs.get("base_dt", int(timedelta(days=1).total_seconds()) * SECONDS_TO_MICROSECONDS_FACTOR))

    def set_start_date(self, start_date: date) -> RollingWindowTimeAxisBuilder:
        if (start_date is None) or isinstance(start_date, date):
            self._start_date = start_date
        else:
            raise TypeError("start_date must be of type date")

        return self

    def set_end_date(self, end_date: date) -> RollingWindowTimeAxisBuilder:
        if (end_date is None) or isinstance(end_date, date):
            self._end_date = end_date
        else:
            raise TypeError("end_date must be of type date")
        return self

    def set_n_window(self, n_window: int) -> RollingWindowTimeAxisBuilder:
        if isinstance(n_window, int):
            if n_window > 0:
                self._n_window = n_window
            else:
                raise ValueError("n_window must be a non-zero and positive integer.")
        elif n_window is None:
            self._n_window = None
        else:
            raise TypeError("n_window must be an int")

        return self

    def set_window_size(self, window_size: int):
        if isinstance(window_size, int):
            if (window_size > 0) or (window_size % 2 != 1):
                self._window_size = window_size
            else:
                raise ValueError("window_size must be an odd positive number.")
        elif window_size is None:
            self._window_size = None
        else:
            raise TypeError("window_size must be an int")

        return self

    def set_base_dt(self, base_dt: int):
        if isinstance(base_dt, int):
            if base_dt > 1:
                self._base_dt = base_dt
            else:
                raise ValueError("base_dt must be a positive integer.")
        elif base_dt is None:
            self._base_dt = None
        else:
            raise TypeError("base_dt must be an int")

        return self

    def prebuild_check(self) -> (bool, Exception):
        if self._start_date is None:
            raise ValueError("start_date is not provided.")

        if self._base_dt is None:
            raise ValueError("Some how base_dt ended up to be None. It cannot be None")

        if self._window_size is None:
            raise ValueError("Window_size is not provided. window_size must a positive integer.")

        if (self._n_window is not None) and (self._end_date is not None):
            raise ValueError("You could provide either the end_date or the n_window; but not both.")

        if (self._n_window is None) and (self._end_date is None):
            raise ValueError("Neither end_date nor the n_window is provided. "
                             "You must provide exactly one of them.")

        if (self._start_date is not None) and (self._end_date is not None) and (self._start_date > self._end_date):
            raise ValueError("start_date must be before end_date.")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            if self._end_date is not None:
                self._n_window = np.ceil(
                    (TimeAxisBuilder.datetime_to_timestamp(self._end_date) -
                     TimeAxisBuilder.datetime_to_timestamp(self._start_date)) / self._base_dt
                ) - (self._window_size - 1)
                if self._n_window < 1:
                    raise ValueError("the provided end_date and start_date resulted in 0 n_window.")

            lower_bound = TimeAxisBuilder.datetime_to_timestamp(self._start_date) + \
                          np.arange(self._n_window, dtype="int64") * self._base_dt

            window_length = self._window_size * self._base_dt
            upper_bound = lower_bound + window_length
            data_tick = 0.5 * (lower_bound + upper_bound)
            return Axis(
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                data_ticks=data_tick
            )


class MonthlyTimeAxisBuilder(TimeAxisBuilder):
    def __init__(self, start_year: int, end_year: int, start_month: int =1, end_month: int = 12):
        if (start_year is not None) and (start_month is not None):
            self.set_start_year_month(start_year, start_month)
        else:
            self._start = None

        if (end_year is not None) and (end_month is not None):
            self.set_end_year_month(end_year, end_month)
        else:
            self._end = None

    def set_start_year_month(self, start_year: int, start_month: int = 1) -> MonthlyTimeAxisBuilder:
        tmp_start_year = int(start_year)
        tmp_start_month = int(start_month)

        if 1<= start_month <= 12:
            self._start = date(tmp_start_year, tmp_start_month, 1)
        else:
            raise ValueError("start_year/month must be convertible to an integer value and "
                             "start_month must be a number between 1 and 12")

        return self

    def set_end_year_month(self, end_year: int, end_month: int = 12) -> MonthlyTimeAxisBuilder:
        tmp_end_year = int(end_year)
        tmp_end_month = int(end_month)

        if 1 <= end_month <= 12:
            self._end = date(tmp_end_year, tmp_end_month, monthrange(tmp_end_year, tmp_end_month)[1])
        else:
            raise ValueError("end_year/month must be convertible to an integer value and "
                             "end_month must be a number between 1 and 12")

        return self

    def prebuild_check(self) -> (bool, Exception):
        if (self._start is None) or (self._end is None):
            raise ValueError("start and/or end year/month is not provided")

        if self._end < self._start:
            raise ValueError("start year/month must be before end year/month")

        return True

    def build(self) -> Axis:
        if self.prebuild_check():
            n_month_in_first_year = 12 - self._start.month + 1
            n_month_in_last_year = self._end.month
            n_years_in_between = self._end.year - self._start.year - 1

            n_months = n_month_in_first_year + n_years_in_between * 12 + n_month_in_last_year

            year_list = np.concatenate((
                np.ones(n_month_in_first_year, dtype="int") * self._start.year,
                np.repeat(np.arange(n_years_in_between, dtype="int") + self._start.year + 1, 12),
                np.ones(n_month_in_last_year, dtype="int") * self._end.year
            ))
            month_list = (np.arange(n_months, dtype="int") + (self._start.month - 1)) % 12 + 1
            lower_bound = TimeAxisBuilder.datetime_to_timestamp(
                [date(y, m, 1) for y, m in zip(year_list, month_list)]
            ).reshape((-1, ))
            upper_bound = TimeAxisBuilder.datetime_to_timestamp(
                [date(y, m, monthrange(y, m)[1]) for y, m in zip(year_list, month_list)]
            ).reshape((-1, ))

            data_ticks = 0.5 * (lower_bound + upper_bound)

            return Axis(
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                data_ticks=data_ticks
            )




from enum import Enum
from numbers import Number


class AxisBinding(Enum):
    BEGINNING = ("beginning", 0.0)
    END = ("end", 1.0)
    MIDDLE = ("middle", 0.5)
    CUSTOM_FRACTION = ("custom_fraction", None)

    def __str__(self):
        return self.value[0]

    def __repr__(self):
        return self.__str__()

    def fraction(self):
        return self.value[1]

    @staticmethod
    def valueOf(v):
        if isinstance(v, str):
            possibleCases = {
                "beginning": AxisBinding.BEGINNING,
                "lef": AxisBinding.BEGINNING,
                "end": AxisBinding.END,
                "right": AxisBinding.END,
                "middle": AxisBinding.MIDDLE,
                "center": AxisBinding.MIDDLE,
                "custom": AxisBinding.CUSTOM_FRACTION,
                "custom_fraction": AxisBinding.CUSTOM_FRACTION
            }
            lowerV = v.lower()
            if lowerV in possibleCases:
                return possibleCases[lowerV]
            else:
                raise ValueError(f"could not find the Time Axis Binding associated to {str(v)}. "
                                  f"possible options are [{', '.join(possibleCases.keys())}]")
        if isinstance(v, Number):
            if v == 0.0:
                return AxisBinding.BEGINNING
            if v == 1.0:
                return AxisBinding.END
            if v == 0.5:
                return AxisBinding.MIDDLE
            if (v >= 0.0) and (v <= 1.0):
                return AxisBinding.CUSTOM_FRACTION
            else:
                raise ValueError("a number between 0 and 1 must be provided.")
        elif isinstance(v, AxisBinding):
            return v
        else:
            raise TypeError("Either an str or TimeAxisBinding object is accepted.")




